/*
 * mode-menu.c
 *
 * Copyright 1998, 1999 Michael Elizabeth Chastain, <mailto:mec@shout.net>.
 * Licensed under the Gnu Public License, version 2.
 *
 * This is the second full-screen-character-mode program I ever wrote!
 * So if you notice that you know something about curses that I don't,
 * feel free to instruct me.
 */

#include <ctype.h>
#include <curses.h>
#include <stdlib.h>
#include <unistd.h>

#include "mconfig.h"
#include "parser.tab.h"



/*
 * All the windows.  These are nonoverlapping whole-line windows.
 * I could scope some of these more locally, but I find it more clear
 * to list them all in one place.
 *
 * FIXME: wmore, wless deserve their own window.
 */

WINDOW * wtitle;	/* very top line, "Linux Kernel Configuration"	*/
WINDOW * wname;		/* name of current menu				*/
WINDOW * wbody;		/* body of current menu (most of the screen)	*/
WINDOW * wstatus;	/* status line at bottom			*/



/*
 * Menu state.
 *
 * I keep track of menus with a classical menu stack.  I implement the
 * 'help' menu by pushing synthetic menus onto the stack.
 *
 * The major data structure here is a dynamically sized array of lines
 * and several associated indices.  Each line represents one window line
 * (not one statement).  The line points to the statement which generated
 * it.
 *
 * 'aline' is the actual array; 'sline' is the maximum size for memory
 * management purposes.  'aline' grows dynamically as needed, of course.
 *
 * During a render-walk, 'nline' is the index of the next line.  After
 * rendering is done, 'nline' is the high bound of the actual lines
 * rendered.  This varies as the user enables and disables options,
 * putting more or fewer sub-options on the window.
 *
 * 'hline' is the currently highlighted line.  Its range is [0..nline).
 * Like 'nline', it is relative to 0 at the beginning of the menu.
 *
 * 'wline' is the size of the window minus two (so, typically, LINES-5).
 * The 'minus two' is for the '--less--' and '--more--' indicators.
 * These indicators perhaps should go in separate windows.  'wline' changes
 * only if the window size changes (maybe I'll write a SIGWINCH handler
 * someday).
 *
 * A menu can be larger than a display window: in other words, 'nline'
 * can be larger than 'wline'.  That's where 'fline' comes in.  It is the
 * first line of the menu that is displayed in the window: the menu
 * line number that corresponds to window line 0.  So the window shows
 * menu lines in the range '[fline, min(fline+wline, nline))'.
 *
 * 'hline' is almost always in the visible window.  So far, the only
 * time that it's not is when 'nline' is zero, no visible lines.  Watch
 * out for these cases; they have caused problems in the past.
 *
 * By the way, those '[..)' unbalanced delimiters are standard notation
 * for a half-open range (the usual notion of range in C).  If you
 * haven't seen this notation before, get used to it.  :)
 *
 * CURSOR MOVEMENT AND SCROLLING
 *
 * If the user hits downarrow, 'hline' increments by one, unless it
 * bumps into 'nline'.  If the highlight is on the last line of the window,
 * then 'hline' is equal to 'fline+wline-1'; so when I increment 'hline',
 * I also increment 'fline' and the window scrolls down.
 *
 * Similarly, if the user hits uparrow, 'hline' decrements by one,
 * unless it bumps into line 0.  If the highlight is on the first line of
 * the window, then 'hline' is equal to 'fline'; so when I decrement 'hline',
 * I also decrement 'fline' and the window scrolls up.
 */

typedef struct mstate_tag
{
    struct mstate_tag *		parent;	/* parent mstate		*/
    const statement_type *	menu;	/* menu statement		*/
    int				ishelp;	/* this is the top help menu	*/
    int				wline;	/* size of window		*/
    const statement_type **	aline;	/* array of window lines	*/
    int				sline;	/* size of aline		*/
    int				nline;	/* index of current line	*/
    int				hline;	/* index of highlighted line	*/
    int				fline;	/* first line in window		*/
    int				icol;	/* current column		*/
} mstate_type;



/*
 * Render one statement.  This returns the number of lines rendered,
 * which could be:
 *
 *    0   assignment statement, disabled dep statement,
 *            any statement inside a dead conditional block
 *    1   bool, tristate
 *    2   hex, int, string
 *    *   choice list
 *
 * This function figures out the correct line count.  No other function
 * knows or depends on the above information; they all use the return
 * value from this function.
 *
 * During the rendering, I compute the column location for the cursor.
 * I can't render it because it doesn't stick (it has to be the last
 * thing rendered).  If this statement includes the highlighted line,
 * then I store this value into a caller-provided location.
 *
 * Parameters:
 *   'wline'    # of lines in the window (probably LINES-5)
 *   'rline'    window line to render on
 *   'hline'    highlight line
 *
 * The caller takes care of all the 'fline' offsets.  'rline' is relative
 * to the beginning of the window (actually, it still needs an offset
 * for the "--less--" indicator).  'hline' is relative to the lines this
 * statement will generate.  So 'hline==0' means this line is highlighted;
 * and 'hline==1' will match the second line of an 'int' command.
 *
 * Note that even if 'rline' is out of range, some statements have
 * side effects (assignment statements), and some statements can display
 * multiple lines and thus can partially appear at the top of the window.
 * Even statements past the bottom need to be rendered for the sake of
 * the "--more-- (%d line%s)" indicator.  Of course, rendering statements
 * outside the window is very fast, because all I need to do is count
 * lines properly.
 */
static int statement_render( const statement_type * statement,
    int wline, int rline, int hline, int * icol )
{
    switch ( statement->verb )
    {
    default:
	error_enum_bogus( );
	return 0;

    case verb_MENU:
	if ( rline >= 0 && rline < wline )
	{
	    if ( hline == 0 )
	    {
		*icol = 1;
		wattron( wbody, A_STANDOUT );
	    }

	    mvwaddnstr( wbody, 1 + rline, 0, "--> ", 4 );
	    waddnstr( wbody,
		statement->sc_title->value.ptr,
		statement->sc_title->value.len );

	    if ( hline == 0 )
		wattroff( wbody, A_STANDOUT );
	}
	return 1;

    case verb_mainmenu_name:
	/* ignore this statement */
	return 0;

    case verb_comment:
	if ( rline >= 0 && rline < wline )
	{
	    piece_type prompt = statement->sb_prompt->value;

	    if ( hline == 0 )
	    {
		*icol = 4;
		wattron( wbody, A_STANDOUT );
	    }

	    mvwaddnstr( wbody, 1 + rline, 0, "--- ", 4 );
	    waddnstr( wbody, prompt.ptr, prompt.len );

	    if ( hline == 0 )
		wattroff( wbody, A_STANDOUT );
	}
	return 1;

    case verb_text:
	if ( rline >= 0 && rline < wline )
	{
	    piece_type prompt = statement->sb_prompt->value;

	    if ( hline == 0 )
	    {
		*icol = 0;
		wattron( wbody, A_STANDOUT );
	    }

	    mvwaddnstr( wbody, 1 + rline, 0, prompt.ptr, prompt.len );

	    if ( hline == 0 )
		wattroff( wbody, A_STANDOUT );
	}
	return 1;

    case verb_unset:
	{
	    symbol_type * symbol;

	    for ( symbol = statement->sb_symbol; symbol != NULL; symbol = symbol->next_list )
	    {
		symbol->value = piece_empty;
		symbol->origin = origin_unset;
	    }
	}
	return 0;

    case verb_def_bool:
    case verb_def_hex:
    case verb_def_int:
    case verb_def_string:
    case verb_def_tristate:
	statement->sb_symbol->value  = word_eval( statement->sb_def_value, 1 );
	statement->sb_symbol->origin = origin_statement;
	return 0;

    case verb_ask_bool:
    case verb_dep_bool:
	if ( ( statement_bools_allowed( statement ) & ~0x01 ) == 0 )
	    return 0;

	if ( rline >= 0 && rline < wline )
	{
	    piece_type prompt = statement->sb_prompt->value;
	    piece_type value  = statement->sb_symbol->value;

	    if ( hline == 0 )
	    {
		*icol = 1;
		wattron( wbody, A_STANDOUT );
	    }

	    if      ( value.len == 1 && value.ptr [0] == 'y' )
	    {
		mvwaddnstr( wbody, 1 + rline, 0, "[*] ", 4 );
	    }
	    else
	    {
		mvwaddnstr( wbody, 1 + rline, 0, "[ ] ", 4 );
		/* force value to be 'n', whatever it was before */
		statement->sb_symbol->value = piece_n;
	    }

	    waddnstr( wbody, prompt.ptr, prompt.len );
	    if ( statement->sb_symbol->origin == origin_unset )
		waddstr( wbody, " (NEW)" );

	    if ( hline == 0 )
		wattroff( wbody, A_STANDOUT );
	}
	return 1;

    case verb_ask_tristate:
    case verb_dep_tristate:
	if ( ( statement_bools_allowed( statement ) & ~0x01 ) == 0 )
	    return 0;

	if ( rline >= 0 && rline < wline )
	{
	    piece_type prompt = statement->sb_prompt->value;
	    piece_type value  = statement->sb_symbol->value;

	    if ( hline == 0 )
	    {
		*icol = 1;
		wattron( wbody, A_STANDOUT );
	    }

	    if      ( value.len == 1 && value.ptr [0] == 'y' )
		mvwaddnstr( wbody, 1 + rline, 0, "<*> ", 4 );
	    else if ( value.len == 1 && value.ptr [0] == 'm' )
		mvwaddnstr( wbody, 1 + rline, 0, "<m> ", 4 );
	    else
	    {
		mvwaddnstr( wbody, 1 + rline, 0, "< > ", 4 );
		/* force value to be 'n', whatever it was before */
		statement->sb_symbol->value = piece_n;
	    }

	    waddnstr( wbody, prompt.ptr, prompt.len );
	    if ( statement->sb_symbol->origin == origin_unset )
		waddnstr( wbody, " (NEW)", 6 );

	    if ( hline == 0 )
		wattroff( wbody, A_STANDOUT );
	}
	return 1;

    case verb_ask_hex:    case verb_dep_hex:
    case verb_ask_int:    case verb_dep_int:
    case verb_ask_string: case verb_dep_string:
	if ( ( statement_bools_allowed( statement ) & ~0x01 ) == 0 )
	    return 0;

	{
	    piece_type prompt = statement->sb_prompt->value;
	    piece_type value  = statement->sb_symbol->value;

	    /* first the header line */
	    if ( rline >= 0 && rline < wline )
	    {
		if ( hline == 0 )
		{
		    *icol = 4;
		    wattron( wbody, A_STANDOUT );
		}

		mvwaddnstr ( wbody, 1 + rline, 0, "... ", 4 );
		waddnstr   ( wbody, prompt.ptr, prompt.len );

		if ( statement->sb_symbol->origin == origin_unset )
		{
		    waddstr( wbody, " (NEW)" );
		    /* this should be a new kind of origin */
		    value = word_eval( statement->sb_def_value, 0 );
		    statement->sb_symbol->value = value;
		}

		if ( hline == 0 )
		    wattroff( wbody, A_STANDOUT );
	    }
	    rline++;

	    /* then the data line */
	    if ( rline >= 0 && rline < wline )
	    {
		mvwaddnstr( wbody, 1 + rline, 0, "      ", 6 );

		if ( hline == 1 )
		{
		    *icol = 6 + value.len;
		    if ( statement->verb == verb_ask_hex
		    ||   statement->verb == verb_dep_hex )
			*icol += 2;
		    if ( statement->verb == verb_ask_string
		    ||   statement->verb == verb_dep_string )
			*icol += 1;
		    wattron( wbody, A_STANDOUT );
		}

		if ( statement->verb == verb_ask_hex
		||   statement->verb == verb_dep_hex )
		    waddnstr( wbody, "0x", 2 );

		if ( statement->verb == verb_ask_string
		||   statement->verb == verb_dep_string )
		    waddch( wbody, '"' );

		waddnstr ( wbody, value.ptr, value.len );

		if ( statement->verb == verb_ask_string
		||   statement->verb == verb_dep_string )
		    waddch( wbody, '"' );

		if ( hline == 1 )
		    wattroff( wbody, A_STANDOUT );
	    }
	}
	return 2;

    case verb_nchoice:
	{
	    const prompt_type * prompt;
	    symbol_type * symbol;
	    int rline_original = rline;
	    int is_new = 0;

	    /*
	     * Count the header line.  I can't render it yet because
	     * its " (NEW)" status depends on the pedigrees of all
	     * the lines below it.
	     */
	    rline++;

	    /*
	     * Render subordinate lines.
	     */
	    for ( prompt  = statement->sn_ps_list->prompt_first,
	          symbol  = statement->sn_ps_list->symbol_first;
		  prompt != NULL && symbol != NULL;
		  prompt  = prompt->next, symbol = symbol->next_list )
	    {
		if ( symbol->origin == origin_unset )
		    is_new = 1;

		if ( rline >= 0 && rline < wline )
		{
		    mvwaddnstr( wbody, 1 + rline, 0, "    ", 4 );
		    if ( rline - rline_original == hline )
		    {
			*icol = 5;
			wattron( wbody, A_STANDOUT );
		    }

		    if ( symbol->value.len == 1 && symbol->value.ptr [0] == 'y' )
			waddnstr( wbody, "[*] ", 4 );
		    else
		    {
			waddnstr( wbody, "[ ] ", 8 );
			/* force value to be 'n', whatever it was before */
			symbol->value = piece_n;
		    }

		    waddnstr( wbody, prompt->value.ptr, prompt->value.len );

		    if ( rline - rline_original == hline )
			wattroff( wbody, A_STANDOUT );
		}

		rline++;
	    }

	    /*
	     * Now render the header line.
	     */
	    if ( rline_original >= 0 && rline_original < wline )
	    {
		if ( hline == 0 )
		{
		    *icol = 4;
		    wattron( wbody, A_STANDOUT );
		}

		mvwaddnstr ( wbody, 1 + rline_original, 0, "### ", 4 );
		waddnstr   ( wbody,
		    statement->sn_prompt->value.ptr,
		    statement->sn_prompt->value.len );
		if ( is_new )
		    waddstr( wbody, " (NEW)" );

		if ( hline == 0 )
		    wattroff( wbody, A_STANDOUT );
	    }
	    return rline - rline_original;
	}
	break;

    }
}




/*
 * Re-render a statement for cursor movement.  I may have 'iline' pointing
 * into the middle of a multi-line statement, so I have to back up to the
 * beginning of the statement.
 */
static void statement_rerender( const mstate_type * mstate,
    int iline, int * icol )
{
    while ( iline > 0 && mstate->aline [iline-1] == mstate->aline [iline] )
	iline = iline-1;

    statement_render( mstate->aline [iline], mstate->wline,
	iline - mstate->fline, mstate->hline - iline, icol );
}



/*
 * Render one statement and then update the mstate line pointers.
 * 'dline' is the number of lines displayed by this statement.
 */
static void scb_render( const statement_type * statement, void * param_mstate )
{
    mstate_type * mstate = (mstate_type *) param_mstate;
    int dline;

    /* don't render the first comment in a block */
    /* FIXME: this fails to render all initial comments! */
    if ( mstate->nline == 0 && statement->verb == verb_comment )
	return;

    /* render the statement */
    dline = statement_render( statement, mstate->wline,
	mstate->nline - mstate->fline, mstate->hline - mstate->nline,
	& mstate->icol );

    /* expand aline as needed */
    while ( mstate->nline + dline >= mstate->sline )
    {
	mstate->aline = check_realloc( mstate->aline,
	    (mstate->sline *= 2) * sizeof(*mstate->aline) );
    }

    /* fill out all the aline entries for this statement */
    {
	int iline = mstate->nline;
	mstate->nline += dline;
	for ( ; iline < mstate->nline; iline++ )
	    mstate->aline [iline] = statement;
    }
}



/*
 * This function belongs somewhere else when it settles down.
 */

extern const char str_help [];

const statement_type * statement_create_help( )
{
    static statement_type * statement_help = NULL;

    if ( statement_help == NULL )
    {
	input_push_string( str_help, strlen(str_help) );
	parser_magic_cookie = COOKIE_HELP_TEXT;
	yyparse( );
	statement_help = parser_block_top->first;

	/* pointer smashing fun! */
	statement_help->sc_title->value.ptr = "Mconfig Help Menu";
	statement_help->sc_title->value.len = 17;
    }

    return statement_help;
}




/*
 * Show the title window.
 */
static void window_wtitle( const char * title )
{
    werase( wtitle );
    mvwaddstr( wtitle, 0, 0, title );
    wnoutrefresh( wtitle );
}



/*
 * Show the name of this menu.
 */
static void window_wname( const mstate_type * mstate )
{
    piece_type title = mstate->menu->sc_title->value;

    werase       ( wname );
    wmove        ( wname, 0, 0 );
    waddnstr     ( wname, title.ptr, title.len );
    wnoutrefresh ( wname );
}



/*
 * Show a status message.
 * The tricky part is restoring the cursor.
 */
static void window_wstatus( const char * message )
{
    int ybody, xbody;
    getyx( wbody, ybody, xbody );

    werase( wstatus );
    if ( message != NULL )
    {
	mvwaddstr( wstatus, 0, 0, message );
	beep( );
    }
    wnoutrefresh( wstatus );

    wmove( wbody, ybody, xbody );
    wnoutrefresh( wbody );

    doupdate( );
}



/*
 * Universal commands.
 * Returns:
 *  -1  did not handle the command
 *   0  handled the command but did not redraw windows
 *   1  handled the command and redrew windows
 */
static int command_universal( int c, mstate_type ** pmstate,
    const block_type * block )
{
    mstate_type * mstate = * pmstate;

    switch ( c )
    {
    default:
	return -1;

    case 0x0C:
    case 0x12:
	/*
	 * Physical screen repaint.
	 */
	clearok  ( wbody, TRUE );
	wrefresh ( wbody );
	return 1;

    case 'Q':
	/* no saving throw */
	endwin( );
	exit( 0 );

    case KEY_F(1):
    case 'h':
	{
	    mstate_type * mstate_help;

	    /*
	     * Look for an active help menu.
	     */
	    for ( mstate_help  = mstate;
	          mstate_help != NULL;
		  mstate_help  = mstate_help->parent )
	    {
		if ( mstate_help->ishelp )
		    break;
	    }

	    if ( mstate_help != NULL )
	    {
		while ( mstate != mstate_help )
		{
		    mstate_type * mstate_free = mstate;
		    mstate = mstate->parent;
		    free( mstate_free->aline );
		    free( mstate_free );
		}
	    }
	    else
	    {
		mstate         = check_malloc( sizeof(*mstate) );
		mstate->parent = * pmstate;
		mstate->menu   = statement_create_help( );
		mstate->ishelp = 1;
		mstate->wline  = mstate->parent->wline;
		mstate->aline  = check_malloc( 64 * sizeof(*mstate->aline) );
		mstate->sline  = 64;
		mstate->nline  = 0;
		mstate->hline  = 0;
		mstate->fline  = 0;
	    }

	    * pmstate = mstate;
	    window_wname( mstate );
	    return 0;
	}

    case 'H':
    case KEY_HOME:
	/*
	 * Warp to top of menu.
	 * I could optimize this, especially if fline == 0.
	 */
	mstate->hline = 0;
	mstate->fline = 0;
	return 0;

    case 'L':
    case KEY_END:
	/*
	 * Warp to bottom of menu.
	 */
	if ( mstate->nline > 0 )
	{
	    mstate->hline = mstate->nline - 1;
	    if ( mstate->fline < mstate->hline - (mstate->wline - 1) )
		mstate->fline = mstate->hline - (mstate->wline - 1);
	}
	return 0;

    case '=':
    case KEY_IC:
    case KEY_DC:
	/*
	 * Move to mid screen.
	 */
	{
	    int lastline = mstate->fline + mstate->wline;
	    if ( lastline > mstate->nline )
		lastline = mstate->nline;
	    mstate->hline = (mstate->fline + lastline) / 2;
	}
	return 0;

    case '+':
    case KEY_NPAGE:
	/*
	 * I messed with this until the results seemed intuitive to me.
	 * If hline is not on the last line of the screen, move there.
	 * If hline is on the last line of the screen, keep the same
	 *   hline, but warp fline so that hline is first line of the
	 *   next screen.
	 */

	/* obviously can't go anywhere from here */
	if ( mstate->hline >= mstate->nline - 1 )
	{
	    window_wstatus( "You are already on the last line." );
	    return 1;
	}

	if ( mstate->hline < mstate->fline + mstate->wline - 1 )
	{
	    mstate->hline = mstate->fline + mstate->wline - 1;
	    if ( mstate->hline > mstate->nline - 1 )
		mstate->hline = mstate->nline - 1;
	}
	else
	{
	    mstate->fline = mstate->hline;
	}
	return 0;

    case '-':
    case KEY_PPAGE:
	/*
	 * Like NPAGE: either hline gets improved, or hline stays
	 * the same and fline decrements as far as it can and keep
	 * hline on the page.
	 */

	/* this is as good as it gets */
	if ( mstate->hline <= 0 )
	{
	    window_wstatus( "You are already on the first line." );
	    return 1;
	}

	if ( mstate->hline > mstate->fline )
	{
	    mstate->hline = mstate->fline;
	}
	else
	{
	    mstate->fline = mstate->hline - (mstate->wline - 1);
	    if ( mstate->fline < 0 )
		mstate->fline = 0;
	}
	return 0;

    case 'k':
    case KEY_UP:
	/*
	 * Cursor up or scroll up.
	 */
	if ( mstate->hline - 1 < 0 )
	{
	    window_wstatus( "You are already on the first line." );
	    return 1;
	}

	mstate->hline = mstate->hline - 1;

	if ( mstate->hline >= mstate->fline )
	{
	    /* cursor up; rerender old and new lines */
	    int icol;
	    statement_rerender( mstate, mstate->hline + 1, &icol );
	    statement_rerender( mstate, mstate->hline,     &icol );
	    mstate->icol = icol;
	    wmove( wbody, 1 + mstate->hline - mstate->fline, mstate->icol );
	    wrefresh( wbody );
	    return 1;
	}
	else
	{
	    /* scroll up; can't render this yet */
	    mstate->fline = mstate->hline;
	    return 0;
	}

    case 'j':
    case KEY_DOWN:
	/* cursor down or scroll down */
	if ( mstate->hline + 1 >= mstate->nline )
	{
	    window_wstatus( "You are already on the last line." );
	    return 1;
	}

	mstate->hline = mstate->hline + 1;

	if ( mstate->hline < mstate->fline + mstate->wline )
	{
	    /* cursor down; rerender old and new lines */
	    int icol;
	    statement_rerender( mstate, mstate->hline - 1, &icol );
	    statement_rerender( mstate, mstate->hline,     &icol );
	    mstate->icol = icol;
	    wmove( wbody, 1 + mstate->hline - mstate->fline, mstate->icol );
	    wrefresh( wbody );
	    return 1;
	}
	else
	{
	    /* scroll down; can't render this yet */
	    mstate->fline = mstate->hline - ( mstate->wline - 1 );
	    return 0;
	}

    case KEY_LEFT:
	if ( mstate->parent == NULL )
	{
	    window_wstatus( "You are already at the top menu." );
	    return 1;
	}
	else
	{
	    * pmstate = mstate->parent;
	    free( mstate->aline );
	    free( mstate );
	    window_wname( * pmstate );
	    return 0;
	}

    case 'W':
	{
	    const char * str_error;

	    if ( block_output( block, &str_error ) == 0 )
	    {
		window_wstatus( "Write successful." );
	    }
	    else
	    {
		char * buf = check_malloc( strlen(str_error) + 16 );
		sprintf( buf, "Write failed: %s", str_error );
		window_wstatus( buf );
		free( buf );
	    }
	}
	return 1;
    }
}



/*
 * Statement-specific commands.
 * Returns:
 *   -1   did not handle command
 *    0   handled command, refreshed all windows
 *    1   handled command, needs refreshing
 */
static int command_statement( int c, mstate_type ** pmstate,
    const block_type * block )
{
    mstate_type * mstate = * pmstate;
    const statement_type * statement;
    
    /*
     * Maybe the highlight isn't anywhere (such as zero-line windows).
     */
    if ( mstate->hline < 0 || mstate->hline >= mstate->nline )
	return -1;

    statement = mstate->aline [mstate->hline];

    switch ( statement->verb )
    {
    default:
	error_enum_bogus( );
	break;

    case verb_MENU:
	switch ( c )
	{
	default:
	    break;

	case ' ':
	case '\r':
	case '\n':
	case KEY_RIGHT:
	    mstate         = check_malloc( sizeof(*mstate) );
	    mstate->parent = * pmstate;
	    * pmstate      = mstate;
	    mstate->menu   = statement;
	    mstate->ishelp = 0;
	    mstate->wline  = mstate->parent->wline;
	    mstate->aline  = check_malloc( 64 * sizeof(*mstate->aline) );
	    mstate->sline  = 64;
	    mstate->nline  = 0;
	    mstate->hline  = 0;
	    mstate->fline  = 0;
	    window_wname( mstate );
	    return 0;
	}
	break;

    case verb_comment:
    case verb_text:
	return -1;

    case verb_ask_bool:     case verb_dep_bool:
    case verb_ask_tristate: case verb_dep_tristate:
	{
	    const int allowed = statement_bools_allowed( statement );

	    switch ( c )
	    {
	    default:
		break;

	    case ' ':
		{
		    int itry;
		    piece_type value = statement->sb_symbol->value;

		    for ( itry = 0; itry < 3; itry++ )
		    {
			int ivalue = 0;

			if ( value.len == 1 )
			{
			    if ( value.ptr [0] == 'n' ) ivalue = 1;
			    if ( value.ptr [0] == 'm' ) ivalue = 2;
			}

			value.ptr = "nmy" + ivalue;
			value.len = 1;

			if ( ( allowed & (1 << ivalue) ) != 0 )
			{
			    statement->sb_symbol->value  = value;
			    statement->sb_symbol->origin = origin_statement;
			    return 0;
			}
		    }

		    window_wstatus( "I can't find any legal value!" );
		    return 1;
		}
		break;

	    case 'N':
	    case 'n':
		if ( ( allowed & 0x01 ) != 0 )
		{
		    statement->sb_symbol->value  = piece_n;
		    statement->sb_symbol->origin = origin_statement;
		    return 0;
		}
		else
		{
		    window_wstatus( "Value 'n' not allowed!" );
		    return 1;
		}
		break;
		
	    case 'M':
	    case 'm':
		if ( ( allowed & 0x02 ) != 0 )
		{
		    statement->sb_symbol->value  = piece_m;
		    statement->sb_symbol->origin = origin_statement;
		    return 0;
		}
		else
		{
		    window_wstatus( "Value 'm' not allowed." );
		    return 1;
		}
		break;

	    case '\n':
	    case '\r':
	    case '*':
	    case 'Y':
	    case 'y':
		if ( ( allowed & 0x04 ) != 0 )
		{
		    statement->sb_symbol->value  = piece_y;
		    statement->sb_symbol->origin = origin_statement;
		    return 0;
		}
		else
		{
		    window_wstatus( "Value 'y' not allowed." );
		    return 1;
		}
		break;
	    }
	}
	break;

    case verb_ask_hex:    case verb_dep_hex:
    case verb_ask_int:    case verb_dep_int:
    case verb_ask_string: case verb_dep_string:
	{
	    piece_type value = statement->sb_symbol->value;
	    int first_line;

	    /* find the first line of this statement */
	    for ( first_line = mstate->hline;
		  first_line > 0 && mstate->aline [first_line-1] == statement;
		  first_line = first_line - 1 )
	    {
		;
	    }

	    if ( first_line == mstate->hline )
		return -1;

	    switch ( c )
	    {
	    default:
		if ( isprint( c ) )
		{
		    if ( ! statement_allow_char( statement, c ) )
		    {
			window_wstatus( "Illegal character for this data type." );
			return 1;
		    }
		    else if ( value.len >= COLS - 16 )
		    {
			window_wstatus( "Value too long." );
			return 1;
		    }
		    else
		    {
			char * ptr_new = grab_memory( value.len + 1 );
			memcpy( ptr_new, value.ptr, value.len );
			ptr_new [value.len++] = c;
			value.ptr = ptr_new;
			statement->sb_symbol->value  = value;
			statement->sb_symbol->origin = origin_statement;
			return 0;
		    }
		}
		break;

	    case '\x08':
	    case '\x7F':
	    case KEY_BACKSPACE:
		if ( value.len == 0 )
		{
		    window_wstatus( "Value is already empty." );
		    return 1;
		}
		else
		{
		    value.len--;
		    statement->sb_symbol->value  = value;
		    statement->sb_symbol->origin = origin_statement;
		    return 0;
		}
	    }
	}
	break;

    case verb_nchoice:
	{
	    symbol_type * symbol;
	    int first_line;

	    /* find the first line of this statement */
	    for ( first_line = mstate->hline;
		  first_line > 0 && mstate->aline [first_line-1] == statement;
		  first_line = first_line - 1 )
	    {
		;
	    }

	    switch ( c )
	    {
	    default:
		break;

	    case ' ':
		{
		    const symbol_type * symbol_yes = NULL;
		    int iline;
		    int yline;

		    /* find the current choice */
		    for ( symbol  = statement->sn_ps_list->symbol_first;
			  symbol != NULL;
			  symbol  = symbol->next_list )
		    {
			piece_type value = symbol->value;
			if ( value.len == 1 && value.ptr [0] == 'y' )
			{
			    symbol_yes = symbol;
			    break;
			}
		    }

		    if ( symbol_yes == NULL )
		    {
			/* use the default symbol */
			symbol_yes = statement->sn_default_symbol;
		    }
		    else
		    {
			/* advance to the next symbol, with wraparound */
			symbol_yes = symbol_yes->next_list;
			if ( symbol_yes == NULL )
			    symbol_yes = statement->sn_ps_list->symbol_first;
		    }

		    /* set the symbols appropriately */
		    yline = 0;
		    for ( symbol  = statement->sn_ps_list->symbol_first, iline = 0;
		          symbol != NULL;
			  symbol  = symbol->next_list, iline++ )
		    {
			symbol->value  = (symbol == symbol_yes) ? piece_y : piece_n;
			symbol->origin = origin_statement;
		    }

		    /* warp hline to the chosen line */
		    if ( mstate->hline != first_line )
		    {
			mstate->hline = first_line + yline;
			if ( mstate->hline  < mstate->fline
			||   mstate->hline >= mstate->fline + mstate->wline )
			    mstate->fline = mstate->hline;
		    }
		}
		return 0;

	    case '\n':
	    case '\r':
	    case '*':
	    case 'Y':
	    case 'y':
		if ( mstate->hline == first_line )
		    return -1;

		for ( symbol  = statement->sn_ps_list->symbol_first;
		      symbol != NULL;
		      symbol  = symbol->next_list )
		{
		    symbol->value  = (++first_line == mstate->hline) ? piece_y : piece_n;
		    symbol->origin = origin_statement;
		}
		return 0;
	    }
	}
	break;

    }

    return -1;
}



/*
 * This is the top-level driver for menu mode.
 */
void do_mode_menu( const block_type * block )
{
    mstate_type *  mstate;
    int is_rendered = 0;

    /* init menu state */
    mstate         = check_malloc( sizeof(*mstate) );
    mstate->parent = NULL;
    mstate->menu   = block->first;
    mstate->ishelp = 0;
    mstate->wline  = 0;
    mstate->aline  = check_malloc( 64 * sizeof(*mstate->aline) );
    mstate->sline  = 64;
    mstate->nline  = 0;
    mstate->hline  = 0;
    mstate->fline  = 0;



    /*
     * Start up curses.
     */
    initscr( );
    nonl( );
    cbreak( );
    noecho( );
    typeahead( -1 );

    wtitle         = newwin(       1, 0,       0, 0 );
    wname          = newwin(       1, 0,       1, 0 );
    wbody          = newwin( LINES-3, 0,       2, 0 );
    wstatus        = newwin(       1, 0, LINES-1, 0 );
    mstate->wline  = LINES-3-2;
    if ( wtitle == NULL || wname == NULL || wbody == NULL || wstatus == NULL )
	error_internal( "do_mode_menu: newwin failed" );

    keypad( wbody, TRUE );



    /*
     * Show title.
     * This happens just once.
     */
    window_wtitle( argument.title );



    /*
     * Show menu name.
     * This happens on each menu change.
     */
    window_wname( mstate );



    /*
     * Get busy in the main loop.
     */
    for ( ; ; )
    {
	if ( is_rendered == 0 )
	{
	    /* show all the body lines */
	    werase( wbody );
	    mstate->nline =  0;
	    mstate->icol  = -1;
	    block_walk( mstate->menu->sc_block_left, 0, scb_render, mstate );
	    if ( mstate->icol < 0 )
	    {
		if ( mstate->nline == 0 )
		    /* oh, no problem */
		    mstate->icol = 1;
		else
		    error_internal( "do_mode_menu: failed icol update" );
	    }

	    /* show the top 'less' indicator */
	    if ( mstate->fline > 0 )
	    {
		mvwprintw( wbody, 0, 0, "--Less-- (%d line%s)",
		    mstate->fline, mstate->fline > 1 ? "s" : "" );
	    }

	    /* show the bottom 'more' indicator */
	    if ( mstate->fline + mstate->wline < mstate->nline )
	    {
		mvwprintw( wbody, 21, 0, "--More-- (%d line%s)",
		    mstate->nline - ( mstate->fline + mstate->wline ),
		    mstate->nline - ( mstate->fline + mstate->wline ) > 1
			? "s" : "" );
	    }

	    /*
	     * Actually show the body
	     * Moving the cursor has to be the last update.
	     */
	    wmove( wbody, 1 + mstate->hline - mstate->fline, mstate->icol );
	    wrefresh( wbody );
	}

	/*
	 * cascading command interpreters
	 */
	{
	    int c = wgetch( wbody );
	    window_wstatus( NULL );
	    is_rendered = command_statement( c, &mstate, block );
	    if ( is_rendered < 0 )
		is_rendered = command_universal( c, &mstate, block );
	    if ( is_rendered < 0 )
	    {
		/* unknown command */
		window_wstatus( "Unknown command (press 'h' for help)." );
		is_rendered = 0;
	    }
	}
    }
}



/*
 * Kludge-o-rama.
 * FIXME: write converter: help.txt -> help.c.
 */

const char str_help [] =

    "# Copyright 1998, 1999 Michael Elizabeth Chastain, <mailto:mec@shout.net>"
    "# Licensed under the Gnu Public License, version 2."
    "#"
    "# text text, glorious text!"
    ""
    "mainmenu_option next_comment"
    "comment 'Release Notes'"
    "  text 'version 0.15'"
    "  text ''"
    "  text ' Linux kernel version 2.2.2 runs split-include just before compiling,'"
    "  text ' rather than just after configuration.  So split-include is completely'"
    "  text ' out of my domain.'"
    "  text ''"
    "  text 'version 0.14'"
    "  text ''"
    "  text ' Tweak the arguments.  In particular, defconfig is gone.'"
    "  text ''"
    "  text ' mconfig reads only one configuration file, not one configuration file'"
    "  text ' on top of one defconfig file.'"
    "  text ''"
    "  text ' add a bunch of def_* and dep_* verbs.'"
    "  text ''"
    "  text 'version 0.13'"
    "  text ''"
    "  text '  More frotz to support 2.0.37-pre4.'"
    "  text ''"
    "  text '  New output.c with full error checking and reporting, .config.old'"
    "  text '  file, and asm symlink.'"
    "  text ''"
    "  text '  --mode minimum and --mode maximum.'"
    "  text ''"
    "  text 'version 0.12'"
    "  text ''"
    "  text '  I have succesfully dogfooded this version -- I built a zImage,'"
    "  text '  although I have not booted the result yet.  I still have to run'"
    "  text '  scripts/split-include by hand.  I am going to incorporate that soon.'"
    "  text ''"
    "  text '  This version has a help system.  Let me know if it is helpful.'"
    "  text ''"
    "  text '  Alan:  --mode syntax is for you.  Check out the x-syntax script.'"
    "  text '  Arjan: --mode random is for you.  Check out the x-random script.'"
    "  text ''"
    "  text '  The screen full of warnings at the start is for real.  I wrote some'"
    "  text '  notes about it at the start of x-syntax.'"
    "endmenu"
    ""
    "mainmenu_option next_comment"
    "comment 'Commands'"
    "  text 'No help available for commands yet.'"
    "endmenu"
    ""
    "mainmenu_option next_comment"
    "comment 'Command Line Options'"
    "  text 'Here are all the mconfig command line options.'"
    "  text ''"
    "  text 'I designed mconfig to be scriptable, so there are options to control'"
    "  text 'all its input files, all its output files, and many other parameters.'"
    "  text 'Many of the options have short names and long names, and the option'"
    "  text 'parser accepts several kinds of abbreviations on top of that.'"
    "  text ''"
    "  text 'For basic interactive use, you do not need any options at all.'"
    "  text 'Sometimes --dir-source, also known as -ds, is convenient.'"
    "  text ''"
    "  text '--help'"
    "  text ''"
    "  text '    Print a list of all the options and exit successfully.'"
    "  text ''"
    "  text '--version'"
    "  text ''"
    "  text '    Print the current version and exit successfully.'"
    "  text ''"
    "  text '-m MODE, --mode MODE'"
    "  text ''"
    "  text '    modes: menu line old x'"
    "  text '    modes: syntax random mutate minimum maximum'"
    "  text ''"
    "  text '    See the Major Modes page for information on the modes.'"
    "  text ''"
    "  text '    Default value: mode --menu.'"
    "  text ''"
    "  text '-a ARCH, --architecture ARCH'"
    "  text ''"
    "  text '    Select the architecture to configure.  This affects the default'"
    "  text '    locations of some input files: arch/$ARCH/Config.in and'"
    "  text '    arch/$ARCH/defconfig.'"
    "  text ''"
    "  text '    Default value: the native architecture of your machine.'"
    "  text ''"
    "  text '--seed SEED'"
    "  text ''"
    "  text '    Specify the random number generator seed for those modes that use'"
    "  text '    random numbers.  If you do not specify a seed, mconfig picks its'"
    "  text '    own seed.  Whatever its source, mconfig reports the seed value in'"
    "  text '    its output files.'"
    "  text ''"
    "  text '    Default value: something random.'"
    "  text ''"
    "  text '--title TITLE'"
    "  text ''"
    "  text '    Specify the title for this configuration.  Menu mode shows the'"
    "  text '    title on the first line of the screen.  Also, mconfig writes the'"
    "  text '    title into the output files.'"
    "  text ''"
    "  text '    Default value: mconfig makes a nice default title including the'"
    "  text '    architecture name and the kernel version number (which it gets by'"
    "  text '    reading the top-level Makefile).'"
    "  text ''"
    "  text '--ds DIR, --dir-source DIR'"
    "  text ''"
    "  text '    Specify the directory of the kernel source.'"
    "  text ''"
    "  text '    Default value: the current directory.'"
    "  text ''"
    "  text '--dt DIR, --dir-target DIR'"
    "  text ''"
    "  text '    Specify the target directory, where mconfig writes all its output'"
    "  text '    files.  mconfig reads all its input files from the source tree and'"
    "  text '    writes all its output files to the target tree.  So if you use this'"
    "  text '    option, you can make configuration files for a read-only source'"
    "  text '    tree.  (You cannot actually build a kernel this way yet; that'"
    "  text '    will require some Makefile changes).'"
    "  text ''"
    "  text '    Default value: the same as $dir-source'"
    "  text ''"
    "  text '--fs FILE, --file-script FILE'"
    "  text ''"
    "  text '    Specify the script file.  This is a file which defines the options'"
    "  text '    and menus you see during configuration.  Usually, you just use the'"
    "  text '    provided script file for your architecture.'"
    "  text ''"
    "  text '    Default value: $dir-source/$arch/config.in'"
    "  text ''"
    "  text '--fci FILE, --file-config-in FILE'"
    "  text '--no-fci, --no-file-config-in'"
    "  text ''"
    "  text '    Specify the input name of your .config file, or specify that you'"
    "  text '    have no existing .config file.  If you use --dt, note that although'"
    "  text '    this file is an input file, the default value is in the target'"
    "  text '    tree.  This is because this file is your file, not part of the'"
    "  text '    kernel source.'"
    "  text ''"
    "  text '    Default value: $dir-target/.config.  If this file does not exist,'"
    "  text '    fall back to $dir-source/arch/$ARCH/defconfig.'"
    "  text ''"
    "  text '--fco FILE, --file-config-out FILE'"
    "  text '--no-fco, --no-file-config-out'"
    "  text ''"
    "  text '    Specify the output name for your configuration file.  This is'"
    "  text '    useful when editing configurations which are not the current'"
    "  text '    configuration.'"
    "  text ''"
    "  text '    Default value: $dir-target/.config.  Note that this option does'"
    "  text '    *NOT* default to the input name.'"
    "  text ''"
    "  text '--fac FILE, --file-autoconf-h FILE'"
    "  text '--no-fac, --no-file-autoconf-h'"
    "  text ''"
    "  text '    Specify the output name for the C header output file.  You need'"
    "  text '    this file in order to build the kernel.  You may want to suppress'"
    "  text '    this with --no-fac, but you probably do not want to change this.'"
    "  text ''"
    "  text '    Default value: $dir-target/include/linux/autoconf.h'"
    "  text ''"
    "  text '--no-asm-symlink'"
    "  text ''"
    "  text '    When mconfig writes its output files, it looks for a symbolic link'"
    "  text '    from $dir-target/include/asm to $dir-source/include/asm-$arch.'"
    "  text '    (If $dir-target is the same as $dir-source, which it usually is,'"
    "  text '    the value of the link is simply asm-$arch).  You need this symlink'"
    "  text '    to build the kernel.'"
    "  text ''"
    "  text '    If mconfig does not find the asm symlink with the right value, it'"
    "  text '    creates one.  This option suppresses the creation.'"
    "  text ''"
    "  text '    Default value: make the asm symlink if needed.'"
    "  text ''"
    "  text '--no-backup'"
    "  text ''"
    "  text '    When mconfig writes the configuration file, it makes a backup file'"
    "  text '    with the name ${file-target-config-out}.old, and it uses a'"
    "  text '    temporary file ${file-target-config-out}.new.  You can suppress'"
    "  text '    these files if you want.  Non-interactive modes such as random mode'"
    "  text '    or maximum modes always enable this flag.'"
    "endmenu"
    ""
    "mainmenu_option next_comment"
    "comment 'Major Modes'"
    "  text '    Modes: menu x line old'"
    "  text '    Modes: syntax random mutate minimum maximum'"
    "  text ''"
    "  text '    mconfig has nine major modes: four interactive modes and five'"
    "  text '    non-interactive modes.  Menu mode is the default mode.'"
    "  text ''"
    "  text '    Menu mode brings up a full-screen curses interface.  Most people'"
    "  text '    most of the time will use menu mode.'"
    "  text ''"
    "  text '    X mode is not written yet, not even designed yet.  I am not'"
    "  text '    an X programmer.  If you are, and you want to write a nice X'"
    "  text '    interface for mconfig, e-mail me.'"
    "  text ''"
    "  text '    Line mode is a compatibility mode that looks like the old'"
    "  text '    make configure.  It prompts for options one at a time,'"
    "  text '    with no going back.'"
    "  text ''"
    "  text '    Old mode is like line mode, but it prompts only for options'"
    "  text '    that do not already have values.  It looks like the old'"
    "  text '    make oldconfig.  Old mode is useful if you patch your kernel'"
    "  text '    and you want to be prompted only for new options.'"
    "  text ''"
    "  text '    Syntax mode is a useful non-interactive mode.  It reads all'"
    "  text '    of its input files, writes out errors and warnings, and exits.'"
    "  text '    Syntax mode never writes any files.  In syntax mode, mconfig'"
    "  text '    writes all its diagnostic messages to stdout rather than stderr'"
    "  text '    for your convenience in scripting.  Syntax mode returns an exit'"
    "  text '    code of 0 if the input was clean, or 1 if the input has any'"
    "  text '    errors or warnings.'"
    "  text ''"
    "  text '    Random mode non-interactively generates a random configuration.'"
    "  text '    You can provide a seed for the random number generator, or you'"
    "  text '    can let mconfig pick its own seed.  mconfig writes the seed to'"
    "  text '    the output files for your convenience.'"
    "  text ''"
    "  text '    Mutate mode is like random mode.  It starts with an existing'"
    "  text '    configuration and mutates N% of the existing choices.  You'"
    "  text '    specify the mutation percentage and the optional seed.'"
    "  text ''"
    "  text '    Minimum mode answers N to all the boolean and tristate questions'"
    "  text '    and chooses the first option in all choice lists.'"
    "  text ''"
    "  text '    Maximum mode answers Y to all the boolean and tristate questions'"
    "  text '    that it can, falling back to M and N as required.  Maximum mode'"
    "  text '    chooses the last option in all choice lists.'"
    "endmenu"
    ""
    "mainmenu_option next_comment"
    "comment 'History and Credits'"
    "  text 'Copyright 1998 Michael Elizabeth Chastain, <mailto:mec@shout.net>'"
    "  text 'Licensed under the GNU Public License, version 2.'"
    "  text ''"
    "  text 'mconfig is a utility program to configure the linux kernel.'"
    "  text 'It replaces the previous generation of kernel configuration tools:'"
    "  text 'make configure, make oldconfig, make menuconfig, and make xconfig.'"
    "  text ''"
    "  text 'mconfig owes the most to Menuconfig by William Roadcap.  Menuconfig'"
    "  text 'gave me the paradigm of repeatedly walking the script with different'"
    "  text 'verb bindings to perform all the functions.'"
    "  text ''"
    "  text 'Thanks to the all the testers, specifically: Hannu Lyytinen.'"
    "endmenu"
    ""
    "text ''"
    "text 'Quick command list:'"
    "text ''"
    "text ' +  Q   quit'"
    "text ' +  h   help on mconfig (this screen)'"
    "text '    ?   help on a CONFIG_* option'"
    "text '    \\   show source for this menu'"
    "text ''"
    "text '    w   write files (with dialog)'"
    "text ' +  W   write files (no dialog, just do it)'"
    "text ''"
    "text '    r   read new .config file'"
    "text '    D   reset all values to defconfig'"
    "text ''"
    "text 'I have implemented the commands marked with a plus sign.'"

;
